import { RestClient } from "./rest/index.js";
import { WebsocketClient } from "./ws/index.js";
import * as events from "./entities/ws/events.js";
import EventEmitter from "eventemitter3";
import { Dispatch } from "./entities/ws/payload.js";

export interface ClientOptions {
	instance: string;
	token: string;
	debug?: boolean;
}

export type ChatInstance = {
	id: string;
	domain: string;
};

export type ClientEvents = {
	ready: (ready: events.Ready & { instance: ChatInstance }) => void;
	messageCreate: (
		message: events.MessageCreate & { instance: ChatInstance },
	) => void;

	// TODO: have this emit an actual type
	error: (message: string) => void;
};

export default class Client extends EventEmitter<ClientEvents> {
	#options: ClientOptions;

	#restClient: RestClient;
	#wsClient: WebsocketClient;

	constructor(opts: ClientOptions) {
		super();
		this.#options = opts;
		this.#restClient = new RestClient(
			this.#options.instance,
			this.#options.token,
		);
		this.#wsClient = new WebsocketClient(
			this.#options.instance,
			this.#options.token,
			{ debug: this.#options.debug },
		);

		this.#wsClient.on("event", this.#handleEvent);
	}

	get rest() {
		return this.#restClient;
	}

	get ws() {
		return this.#wsClient;
	}

	async connect() {
		await this.#wsClient.connect();
	}

	#handleEvent(instanceId: string, event: Dispatch) {
		const instance = this.#wsClient.instances.find((i) => i.id === instanceId);
		if (!instance) {
			this.emit(
				"error",
				`Unknown instance ID for event with type "${event.t}"`,
			);
			return;
		}

		switch (event.t) {
			case "MESSAGE_CREATE":
				this.emit("messageCreate", { ...event.d, instance });
				break;
			case "READY":
				this.emit("ready", { ...event.d, instance });
		}
	}
}
