import RestError, { ErrorCode } from "./RestError.js";

import type * as ident from "../entities/rest/ident/index.js";
import type * as chat from "../entities/rest/chat/index.js";
import Constants from "../Constants.js";

export default class RestClient {
	#instance: string;
	#token: string;

	constructor(instance: string, token: string) {
		this.#instance = instance;
		this.#token = token;
	}

	/** Returns a new ChatRestClient for the given instance. */
	chat(instance: string) {
		return new ChatRestClient(this, this.#instance, instance);
	}

	/**
	 * Makes a request with this client's token, and handles error responses.
	 * Clients should usually use `requestIdentity` or `requestChat` instead.
	 * @param method The HTTP method to use.
	 * @param url The URL to request.
	 * @param extraHeaders Extra headers to add to the request, if any.
	 * @param body The request body, if any.
	 * @returns A parsed JSON response.
	 * @throws RestError
	 */
	async rawRequest<T>(
		method: string,
		url: string,
		extraHeaders?: Record<string, string>,
		body?: any,
	): Promise<T> {
		const requestBody = body ? JSON.stringify(body) : undefined;
		const headers = requestBody
			? {
					"Content-Type": "application/json",
					"Content-Length": `${requestBody.length}`,
				}
			: undefined;

		const resp = await fetch(url, {
			method,
			headers: {
				...(headers || {}),
				...(extraHeaders || {}),
				Authorization: this.#token,
			},
			body: requestBody,
		});

		// If we didn't get a JSON response, return
		if (resp.headers.get("Content-Type") !== "application/json") {
			throw new RestError(ErrorCode.RESPONSE_NOT_JSON);
		}

		const data = await resp.json();
		if (resp.status < 200 || resp.status >= 400)
			throw new RestError(data.code, data.message);

		return data as T;
	}

	/**
	 * Makes a request to an identity instance endpoint. The string `/_fox/ident` is automatically prepended to the path.
	 * For example, when making a request to `/_fox/ident/node`, `path` should be `/node`.
	 * @param method The HTTP method to use.
	 * @param path The path to request.
	 * @param body The request body, if any.
	 * @returns A parsed JSON response.
	 * @throws RestError
	 */
	request<T>(method: string, path: string, body?: any): Promise<T> {
		return this.rawRequest(
			method,
			`https://${this.#instance}/_fox/ident${path}`,
			undefined,
			body,
		);
	}

	/** Get the current instance's node data. */
	getNode() {
		return this.request<ident.Node>("GET", "/node");
	}
}

export class ChatRestClient {
	#chatInstance: string;
	#identityInstance: string;
	#restClient: RestClient;

	constructor(
		restClient: RestClient,
		identityInstance: string,
		chatInstance: string,
	) {
		this.#restClient = restClient;
		this.#identityInstance = identityInstance;
		this.#chatInstance = chatInstance;
	}

	/**
	 * Makes a request to a proxied chat instance endpoint. The string `/_fox/proxy` is automatically prepended to the path.
	 * For example, when making a request to `/_fox/proxy/channels/01HMKRAG13AGK1WN7XJ0BY2KGB/messages`,
	 * `path` should be `/channels/01HMKRAG13AGK1WN7XJ0BY2KGB/messages`.
	 * @param method The HTTP method to use.
	 * @param path The path to request.
	 * @param body The request body, if any.
	 * @returns A parsed JSON response.
	 * @throws RestError
	 */
	request<T>(method: string, path: string, body?: any): Promise<T> {
		return this.#restClient.rawRequest(
			method,
			`https://${this.#identityInstance}/_fox/proxy${path}`,
			{ [Constants.serverProxyHeader]: this.#chatInstance },
			body,
		);
	}

	createGuild(name: string) {
		return this.request<chat.Guild>("POST", "/guilds", { name });
	}

	createMessage(channelId: string, content: string) {
		return this.request<chat.Message>(
			"POST",
			`/channels/${channelId}/messages`,
			{ content },
		);
	}
}
