import WebSocket from "isomorphic-ws";
import EventEmitter from "eventemitter3";
import {
	ClientPayload,
	Dispatch,
	ServerPayload,
} from "../entities/ws/payload.js";

export type ClientOptions = {
	debug?: boolean;
};

type WebsocketEvents = {
	event: (instanceId: string, event: Dispatch) => void;
	error: (message: string) => void;
	wsError: (error: WebSocket.ErrorEvent) => void;
	wsClose: (event: WebSocket.CloseEvent) => void;
};

export default class Client extends EventEmitter<WebsocketEvents> {
	#instance: string;
	#token: string;
	#ws: WebSocket | undefined;
	#debug = false;
	#connected = false;
	#instances: Array<{ id: string; domain: string }> = [];
	#guilds: string[] = [];

	#heartbeatTimer: any = undefined;
	#heartbeatInterval: number | undefined;
	#lastHeartbeatSend: number | undefined;
	#lastHeartbeatAck: number | undefined;

	constructor(instance: string, token: string, opts?: ClientOptions) {
		super();
		this.#debug = opts?.debug || false;
		this.#instance = instance;
		this.#token = token;
	}

	async connect() {
		this.#debug &&
			console.log(`Connecting to "wss://${this.#instance}/_fox/ident/ws"`);
		this.#ws = new WebSocket(`wss://${this.#instance}/_fox/ident/ws`);

		this.#ws.onerror = (error) => {
			this.emit("wsError", error);
		};

		this.#ws.onopen = () => {
			this.send({ t: "IDENTIFY", d: { token: this.#token } });
		};

		this.#ws.onmessage = (event) => {
			if (typeof event.data === "string") {
				this.handleMessage(JSON.parse(event.data));
			}
		};

		this.#ws.onclose = (event) => {
			this.emit("wsClose", event);
		};
	}

	disconnect() {
		this.#ws?.close();
	}

	send(payload: ClientPayload) {
		if (this.#debug)
			console.log(
				"[CLIENT] Sending %s (%s)",
				payload.t,
				JSON.stringify(payload.d),
			);

		if (!this.#ws) throw "Socket is closed";
		this.#ws.send(JSON.stringify(payload));
	}

	handleMessage(payload: ServerPayload) {
		if (this.#debug)
			console.log(
				"[SERVER] Received %s (%s)",
				payload.t,
				JSON.stringify(payload.d),
			);

		switch (payload.t) {
			case "HELLO":
				this.#connected = true;
				this.#guilds = payload.d.guilds;
				this.#instances = payload.d.instances;
				// Start heartbeat interval
				this.#heartbeatInterval = payload.d.heartbeat_interval;
				if (this.#heartbeatTimer) {
					clearInterval(this.#heartbeatTimer);
				}
				this.#heartbeatTimer = setInterval(() => {
					this.heartbeat();
				}, this.#heartbeatInterval);

				break;
			case "HEARTBEAT_ACK":
				this.#lastHeartbeatAck = Date.now();
				break;
			case "D":
				this.emit("event", payload.d.s, payload.d.e);
				break;
			case "ERROR":
				this.emit("error", payload.d.message);
				break;
		}
	}

	heartbeat() {
		if (
			this.#heartbeatInterval &&
			this.#lastHeartbeatAck &&
			this.#heartbeatInterval * 3 < Date.now() - this.#lastHeartbeatAck
		) {
			this.emit(
				"error",
				`Last heartbeat ACK was more than ${this.#heartbeatInterval * 3}ms ago, disconnecting`,
			);
			this.disconnect();
		}

		this.#lastHeartbeatSend = Date.now();
		this.send({
			t: "HEARTBEAT",
			d: { t: Math.round(this.#lastHeartbeatSend / 1000) },
		});
	}

	get instances() {
		return this.#instances;
	}

	get ping() {
		if (!this.#lastHeartbeatAck || !this.#lastHeartbeatSend) return 0;
		this.#lastHeartbeatAck - this.#lastHeartbeatSend;
	}
}
